from modelgen import Base
from yaml import safe_load, YAMLError

class Helper(Base):

    def __init__(self):
        Base.__init__(self)

    def unpack_kwargs(self, kwargs: dict) -> str:
        """
        Function to format kwargs and translate the dictionary
        to be used to render the jinja template

        Args:
            kwargs (dict): Parameters packed in a python dictionary
                    ex: {'redshift_distkey': 'id', 'redshift_diststyle': 'KEY'}

        Returns:
            (str): ex: redshift_distkey='id',redshift_diststyle='KEY'
        """
        stmt = ','.join((f"{a}={b}" for a, b in kwargs.items()))
        return stmt

    def read_yaml(self, filepath: str) -> dict:
        """
        Function to read yaml and return it's contents in
        the form of python dictionary

        Args:
            filepath (str): path where the yaml file is stored
                            ex: './templates/file.yaml'
        Returns:
            (dict): Python-dictionary representation of the yaml file
        """
        try:
            self.logger.info(f"Reading file {filepath} and converting it to python dict")
            with open(filepath, 'r') as f:
                data = safe_load(f)
            return data
        except YAMLError as e:
            try:
                self.logger.exception(e)
                raise Exception(e)
            finally:
                e = None
                del e

    def write_to_file(self, path, data):
        with open(path, 'w') as f:
            f.write(data)
        return True



import logging, os

class Logging:

    def __init__(self, log_level, filename=None):
        self.log_level = log_level
        self.root_log_dir = 'logs'
        self.filename = 'logs/{}'.format(filename)
        self.logger = logging.getLogger(__name__)

    def get_logger(self):
        logging.basicConfig(level=(self.log_level), format='%(asctime)s - %(message)s',
          datefmt='%d-%b-%y %H:%M:%S')
        return self.logger



from modelgen.log_source import Logging
from os import environ

class Base:

    def __init__(self):
        log_level = environ.get('LOG_LEVEL', 'INFO')
        self.logger = Logging(log_level=log_level).get_logger()



alchemygen = '''from sqlalchemy import (String, Integer, Boolean, Float, 
                                Numeric, DateTime, Date, Table, Column, 
                                PrimaryKeyConstraint, ForeignKey)
from sqlalchemy.ext.declarative import declarative_base

DeclarativeBase = declarative_base()

{{ datasource }}_metadata = DeclarativeBase.metadata\

{% for table, schema in yaml_data[cst.key_tables].items() %}
{% if not bool(schema.get(cst.key_columns)) -%}
    {% set column_data = yaml_data[cst.key_tables][schema.get(cst.inherit_from)][cst.key_columns] %}
{% else -%}
    {% set column_data = schema.get(cst.key_columns) %}
{%- endif %}
{{ table }} = Table('{{ table }}', 
            {{ datasource }}_metadata,{%- for col in column_data -%}
            {%- if bool(col.get(cst.key_column_length)) -%}
                {% set length = col.get(cst.key_column_length) %}
            {%- endif %}
            Column('{{ col[cst.key_column_name] }}'
                    {%- if bool(col.get(cst.key_column_length)) -%}
                        ,{{ cst.sqlalchemy_python_types[col[cst.key_column_type]] }}({{ col.get(cst.key_column_length) }})
                    {%- else -%}
                        ,{{ cst.sqlalchemy_python_types[col[cst.key_column_type]] }}
                    {%- endif %}
                    {%- if bool(col.get(cst.key_foreign_key)) -%}
                        ,ForeignKey('{{ col.get(cst.key_foreign_key) }}')
                    {%- endif %}
                    {%- if cst.key_primary_key in col -%}
                        ,primary_key={{ col.get(cst.key_primary_key) }}
                    {%- endif %}
                    {%- if cst.key_unique in col -%}
                        ,unique={{ col.get(cst.key_unique) }}
                    {%- endif %}
                    {%- if cst.key_nullable in col -%}
                        ,nullable={{ col.get(cst.key_nullable) }}
                    {%- endif %}),
            {%- endfor %}
            {% if bool(schema.get(cst.key_extra_params)) -%}              
                {% for params in schema.get(cst.key_extra_params) -%}
            {{ params[cst.key_extra_params_name] }}={{ params[cst.key_extra_params_value] }},\
                {%- endfor %}
            {%- endif %}             
        )
{% endfor %}
'''

metagen = '''{% set import_list = [] %}\
{% for file in filenames %}\
{% set meta_obj = splitext(file)[0]+'_metadata' %}\
from models.{{ splitext(file)[0] }} import {{ meta_obj }}\
{{ import_list.append(meta_obj)|default("", True)  }}
{% endfor %}
metadata = [{{ import_list|join(',') }}]
'''


from .base import Base
from .helper import Helper
import modelgen.constants
from .templates.alchemygen import alchemygen, metagen
from os import path, getcwd, walk
from jinja2 import Template
from pathlib import Path
from .parser import Parser

def create_model(datasource: str, alembic: bool=False, filepath: str=None) -> bool:
    h = Helper()
    if not filepath:
        filepath = path.join(constants.templates_folder, f"{datasource}.yaml")
    parser = Parser(filepath=filepath)
    src_template = Template(alchemygen)
    py_code = src_template.render(datasource=datasource,yaml_data=parser.data, cst=constants, bool=bool)
    Path(constants.models_folder).mkdir(parents=True, exist_ok=True)
    py_filepath = path.join(constants.models_folder, f'{datasource}.py')
    h.write_to_file(path=py_filepath, data=py_code)
    if alembic:
        create_alembic_meta()
    return True

def create_alembic_meta() -> bool:
    h = Helper()
    alembic_template = Template(metagen)
    _, _, filenames = next(walk(constants.models_folder))
    alembic_meta = alembic_template.render(filenames=filenames, cst=constants,
                                           splitext=path.splitext)
    Path(constants.alembic_meta_folder).mkdir(parents=True, exist_ok=True)
    alembic_meta_filepath = path.join(constants.alembic_meta_folder, '__init__.py')
    h.write_to_file(path=alembic_meta_filepath, data=alembic_meta)
    return True


from modelgen import constants as cst, Helper

class Parser(Helper):

    def __init__(self, filepath: str):
        Helper.__init__(self)
        self.data = self.read_yaml(filepath)
        self.tables = list(self.data[cst.key_tables].keys())
        self.schema = self.get_tables_w_columns()


    def get_tables_w_columns(self) -> dict:
        self.logger.info('Creating schema from YAML')
        schema = dict()
        tables_data = self.data[cst.key_tables]
        for table in self.tables:
            self.logger.debug(f"Getting schema for table {table}")
            column_data = dict()
            if cst.key_inherit_from in tables_data[table]:
                self.logger.debug('Found keyword `inherit_from`')
                self.logger.info(f"Inheriting structure from table {tables_data[table][cst.key_inherit_from]}")
                iter_table = tables_data[tables_data[table][cst.key_inherit_from]][cst.key_columns]
            else:
                self.logger.info(f"Getting structure from table {table}")
                iter_table = tables_data[table][cst.key_columns]
            for column in iter_table:
                column_data.update({column[cst.key_column_name]: column[cst.key_column_type]})

            schema.update({table: column_data})

        return schema



import argparse
from pathlib import Path
from os import path, getcwd
from shutil import copyfile, copytree
from modelgen import __file__, create_model

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i","--init", type=str, help="Initialize sqlalchemygen project")
    parser.add_argument("-c","--create_model", type=str, help="Create sqlalchemy model")
    parser.add_argument("-a", "--alembic", action="store_true", help="If specified, alembic support will be set to True")
    parser.add_argument("-f", "--file", help="path/to/the/template_file.yaml")

    args = parser.parse_args()
    # if args.init and not args.file:
    #     raise TypeError('Please specify folder path. Ex: modelgen init -f ./example')
    # if args.create_model and not args.file:
    #     raise TypeError('Please specify YAML file path. Ex: modelgen create_model -f ./example.yaml')
    try:
        if args.init:
            if path.isabs(args.init):
                f_path = path.join(args.init)
            else:
                f_path = path.join(getcwd(), args.init)
            alembic_path = path.join('/',*(__file__.split('/')[:-1]),'alembic')
            module_src_path = alembic_path
            ini_src_path = path.join('/',*(__file__.split('/')[:-1]),'alembic.ini')
            templates_src_path = path.join('/',*(__file__.split('/')[:-1]),'templates')
            
            templates_dst_path = path.join(args.init, 'templates')
            Path(templates_dst_path).mkdir(parents=True, exist_ok=False)
            Path(path.join(alembic_path,'versions')).mkdir(parents=True, exist_ok=False)

            copytree(module_src_path, path.join(f_path, 'alembic'))
            copyfile(ini_src_path, path.join(f_path, 'alembic.ini'))
            copyfile((path.join(templates_src_path, 'example.yaml')), path.join(templates_dst_path, 'example.yaml'))
        
        elif args.create_model:
            filepath = path.join(args.create_model)
            datasource = filepath.split('.yaml')[0].split('/')[-1]
            create_model(datasource=datasource, filepath=filepath)
            if args.alembic:
                create_model(datasource=datasource, alembic=True, filepath=filepath)
    except FileExistsError as e:
        raise FileExistsError("Folder exists. Please specify a new folder name") from FileExistsError
    

if __name__ == "__main__":
    main()


from os import getcwd, path
key_tables = 'tables'
key_columns = 'columns'
key_column_name = 'name'
key_column_type = 'type'
key_column_length = 'length'
# constraints keys
key_primary_key = 'primary_key'
key_foreign_key = 'foreign_key'
key_unique = 'unique'
key_nullable = 'nullable'

key_extra_params = 'extra_params'
key_extra_params_name = 'name'
key_extra_params_value = 'value'
key_inherit_from = 'inherit_from'

templates_folder = path.join(getcwd(), 'templates')

models_folder = path.join(getcwd(), 'models')
alembic_meta_folder = path.join(getcwd(), 'metadata')

sqlalchemy_python_types = {'smallint':'SmallInteger', 'integer':'Integer',  
                           'bigint':'BigInteger',  'short':'Integer',  'long':'BigInteger', 
                           'decimal':'Numeric',  'numeric':'Numeric',  'double':'Float', 
                           'real':'Float',  'float':'Float',  'boolean':'Boolean', 
                           'char':'String',  'varchar':'String',  'timestamp':'DateTime',
                           'date': 'Date'}



from setuptools import setup, find_packages

setup(
    name = 'alchemy-modelgen',
    url = 'https://github.com/shree14/sqlalchemy-modelgen',
    author = 'Shrinivas Deshmukh',
    author_email = 'shrinivas.deshmukh11@gmail.com',
    version = '0.1.3',
    packages = ['modelgen', 'modelgen.templates', 'modelgen.alembic'],
    install_requires=[
          'alembic>=1.5.8',
          'Jinja2>=2.11.3',
          'mysql-connector-python>=8.0.23',
          'psycopg2-binary>=2.8.6',
          'python-dotenv>=0.17.0',
          'PyYAML>=5.4.1',
          'snowflake-sqlalchemy>=1.2.4',
          'SQLAlchemy>=1.4.4',
          'sqlalchemy-redshift>=0.8.2'
      ],
    setup_requires = [
          'alembic>=1.5.8',
          'Jinja2>=2.11.3',
          'mysql-connector-python>=8.0.23',
          'psycopg2-binary>=2.8.6',
          'python-dotenv>=0.17.0',
          'PyYAML>=5.4.1',
          'snowflake-sqlalchemy>=1.2.4',
          'SQLAlchemy>=1.4.4',
          'sqlalchemy-redshift>=0.8.2'
      ],
    package_data={'modelgen': ['alembic.ini','templates/*', 'alembic/*']},
    entry_points = {
        'console_scripts': [
            'modelgen = modelgen.__main__:main'
        ]
    }
)